/**
 * $Id: yarvcore.h 231 2005-08-15 00:55:30Z ko1 $
 * Create : K.S. 04/01/01 19:41:38
 */

#ifndef _YARVCORE_H_INCLUDED_
#define _YARVCORE_H_INCLUDED_

#include "debug.h"

/*****************/
/* configuration */
/*****************/

/* gcc ver. check */
#if defined(__GNUC__) && __GNUC__ >= 2

#ifdef OPT_TOKEN_THREADED_CODE
#ifdef OPT_DIRECT_THREADED_CODE
#undef OPT_DIRECT_THREADED_CODE
#endif
#endif

#else

/* disable threaded code options */
#ifdef OPT_DIRECT_THREADED_CODE
#undef OPT_DIRECT_THREADED_CODE
#endif
#ifdef OPT_TOKEN_THREADED_CODE
#undef OPT_TOKEN_THREADED_CODE
#endif
#endif

/* call threaded code */
#ifdef OPT_CALL_THREADED_CODE
#ifdef OPT_DIRECT_THREADED_CODE
#undef OPT_DIRECT_THREADED_CODE
#endif
#ifdef OPT_STACK_CACHING
#undef OPT_STACK_CACHING
#endif
#define YARV_AOT_COMPILED 1
#endif

/* likely */
#ifdef __GCC__
#define LIKELY(x)   (__builtin_expect(x, 1))
#define UNLIKELY(x) (__builtin_expect(x, 0))
#else
#define LIKELY(x)   ((x) == 1)
#define UNLIKELY(x) ((x) == 0)
#endif

#define YARVDEBUG 0
#define CPDEBUG   0
#define VMDEBUG   0
#define GCDEBUG   0




/* classes and modules */
extern VALUE mYarvCore;
extern VALUE cYarvISeq;
extern VALUE cYarvVM;
extern VALUE cYarvThread;
extern VALUE mYarvInsns;
extern VALUE cYarvEnv;
extern VALUE cYarvProc;

extern VALUE symIFUNC;
extern VALUE symCFUNC;

/* special id */
extern ID idPLUS;
extern ID idMINUS;
extern ID idMULT;
extern ID idDIV;
extern ID idMOD;
extern ID idLT;
extern ID idLTLT;
extern ID idEq;
extern ID idEqq;
extern ID idBackquote;
extern ID idEqTilde;
extern ID idThrowState;
extern ID idThrowBackDFP;
extern ID idThrowObject;
extern ID idAREF;
extern ID idASET;
extern ID idIntern;
extern ID idMethodMissing;
extern ID idEach;
extern ID idLength;

extern unsigned long yarvGlobalStateVersion;

struct insn_info_struct{
  unsigned short position;
  unsigned short line_no;
};

#define ISEQ_TYPE_TOP    INT2FIX(1)
#define ISEQ_TYPE_METHOD INT2FIX(2)
#define ISEQ_TYPE_BLOCK  INT2FIX(3)
#define ISEQ_TYPE_CLASS  INT2FIX(4)
#define ISEQ_TYPE_RESCUE INT2FIX(5)
#define ISEQ_TYPE_ENSURE INT2FIX(6)

#define CATCH_TYPE_RESCUE INT2FIX(1)
#define CATCH_TYPE_ENSURE INT2FIX(2)
#define CATCH_TYPE_RETRY  INT2FIX(3)
#define CATCH_TYPE_BREAK  INT2FIX(4)
#define CATCH_TYPE_REDO   INT2FIX(5)
#define CATCH_TYPE_NEXT   INT2FIX(6)

struct catch_table_entry{
  VALUE type;
  VALUE iseq;
  unsigned long start;
  unsigned long end;
  unsigned long sp;
  unsigned long cont;
};

#define INITIAL_ISEQ_COMPILE_DATA_STORAGE_BUFF_SIZE (1024)

struct iseq_compile_data_storage{
  struct iseq_compile_data_storage *next;
  unsigned long pos;
  unsigned long size;
  char *buff;
};

struct iseq_compile_data_ensure_node_stack{
  NODE* ensure_node;
  struct iseq_compile_data_ensure_node_stack *prev;
};

struct iseq_compile_data{
  /* GC is needed */
  VALUE err_info;
  VALUE mark_ary;

  /* GC is not needed */
  struct iseq_label_data* start_label;
  struct iseq_label_data* end_label;
  struct iseq_label_data* redo_label;
  VALUE current_block;
  VALUE loopval_popped;  /* used by NODE_BREAK */
  VALUE in_ensure;       /* used by NODE_RETURN */
  VALUE ensure_node;
  struct iseq_compile_data_ensure_node_stack *ensure_node_stack;
  int   cached_const;

  struct iseq_compile_data_storage *storage_head;
  struct iseq_compile_data_storage *storage_current;
};

#define GetISeqVal(obj, tobj) \
  Data_Get_Struct(obj, yarv_iseq_t, tobj)


struct yarv_iseq_struct;

struct yarv_iseq_struct{
  VALUE  name; /* String: iseq name */
  VALUE *iseq; /* iseq */
  VALUE *iseq_encoded;
  VALUE  iseq_mark_ary; /* Array: includes operands which should be GC marked */

  NODE *node;
  
  
  /* sequence size */
  unsigned long size;

  /* insn info, must be freed */
  struct insn_info_struct *insn_info_tbl;
  
  /* insn info size, this value shows also instruction count */
  unsigned int insn_info_size;

  /* file information where this sequence from */
  VALUE file_name;
  
  ID *local_tbl;                  /* must free */
  int local_size;

  /* jit compiled or not */
  void *jit_compiled;
  void *iseq_orig;
  
  
  /**
   * argument information
   *
   *  def m(a1, a2, ..., aM, b1=(...), b2=(...), ..., bN=(...), *c, &d)
   * =>
   *
   *  argc          = M
   *  arg_rest      = M+N
   *  arg_opts      = N
   *  arg_opts_tbl  = [ (N entries) ]
   *  arg_initstart = M+N+1
   *
   *  check:
   *    M <= num
   */
  
  
  int argc;
  int arg_simple;
  int arg_rest;
  int arg_block;
  int arg_opts;
  VALUE *arg_opt_tbl;
  
  /* rewind frame size from dfp */
  int rewind_frame_size;
  
  /* for stack overflow check */
  int stack_max;

  /* instruction sequence type */
  int type;

  /* klass/module nest information stack */
  VALUE klass_nest_stack; /* Array */

  /* catch table */
  VALUE catch_table_ary; /* Array */
  struct catch_table_entry *catch_table;
  int                       catch_table_size;
  
  /* for child iseq */
  struct yarv_iseq_struct *parent_iseqobj;
  struct yarv_iseq_struct *root_iseqobj;
  
  struct iseq_compile_data *compile_data;
};

typedef struct yarv_iseq_struct yarv_iseq_t;



#define GetVMVal(obj, tobj) \
  Data_Get_Struct(obj, yarv_vm_t, tobj)

typedef struct{
  VALUE main_thread;
  VALUE self;
  
  /* this data structure should include all
     runtime information (e.g. method table, object management table, ...) */
} yarv_vm_t;

typedef struct{
  VALUE *pc;                  // cfp[0]
  VALUE *sp;                  // cfp[1]
  VALUE *bp;                  // cfp[2]
  yarv_iseq_t *iseq;          // cfp[3]
  VALUE magic;                // cfp[4]
  VALUE self;                 // cfp[5] // block[0]
  VALUE *lfp;                 // cfp[6] // block[1]
  VALUE *dfp;                 // cfp[7] // block[2]
  yarv_iseq_t * block_iseq;   // cfp[8] // block[3]
  VALUE proc;                 // cfp[9] // block[4]
} yarv_control_frame_t;

typedef struct {
  VALUE self; /* share with method frame if it's only block */
  VALUE *lfp; /* share with method frame if it's only block */
  VALUE *dfp; /* share with method frame if it's only block */
  yarv_iseq_t *iseq;
  VALUE proc;
} yarv_block_t;

#define GetThreadVal(obj, tobj) \
  Data_Get_Struct(obj, yarv_thread_t, tobj)

typedef struct{
  VALUE self;
  VALUE vm_value;
  yarv_vm_t *vm;
  
  /* execution information */
  VALUE *stack;                   /* must free, must mark*/
  unsigned long stack_size;
  yarv_control_frame_t *cfp;

  /* passing state */
  int state;
  
  /* for rb_iterate */
  NODE *ifuncnode;

  /* klass/module nest information stack */
  VALUE klass_nest_stack; /* Array */

  /* other information */
  VALUE stat_insn_usage;
} yarv_thread_t;

/** node -> yarv instruction sequence object */
VALUE iseq_compile(VALUE self, VALUE node);

/** disassemble instruction sequence */
VALUE iseq_disasm(VALUE self);
VALUE iseq_disasm_insn(VALUE str, VALUE *iseq, int pos,
                       yarv_iseq_t *iseqobj, VALUE child);
char* node_name(int node);


/* insns_info.c */
char *insn_name(int insn);
int   insn_len(int insn);
char *insn_op_types(int insn);
int   insn_op_type(int insn, int pos);


/* each thread has this size stack : 2MB */
#define YARV_THREAD_STACK_SIZE 0x20000


/* from ruby 1.9 variable.c */
struct global_entry {
  struct global_variable *var;
  ID id;
};

#define GetProcVal(obj, iobj) \
  Data_Get_Struct(obj, yarv_proc_t, iobj)

typedef struct{
  yarv_block_t block;
  VALUE envval;       /* for GC mark */
  VALUE blockprocval;
} yarv_proc_t;

#define GetEnvVal(obj, iobj) \
  Data_Get_Struct(obj, yarv_env_t, iobj)


typedef struct{
  VALUE *env;
  int    env_size;
  int    local_size;
  VALUE  prev_envval; /* for GC mark */
} yarv_env_t;


/* used by compile time and send insn */
#define VM_CALL_ARGS_SPLAT_BIT    0x01
#define VM_CALL_ARGS_BLOCKARG_BIT 0x02
                         /* 0x04, 0x08 */
#define VM_CALL_TAILCALL_BIT      0x10
#define VM_CALL_TAILRECURSION_BIT      0x20


VALUE thread_invoke_proc_call(VALUE self, VALUE proc, int argc, VALUE *args);
VALUE thread_eval_body(VALUE self);
VALUE iseq_inspect(VALUE self);


/* inline method cache */
#define NEW_INLINE_CACHE_ENTRY() NEW_WHILE(Qundef, 0, 0)
#define ic_klass  u1.value
#define ic_method u2.node
#define ic_value  u2.value
#define ic_vmstat u3.cnt
typedef NODE * IC;

typedef VALUE CDHASH;


#define GC_GUARDED_PTR(p)     ((VALUE)((VALUE)(p) | 0x01))
#define GC_GUARDED_PTR_REF(p) ((void *)(((VALUE)p) & ~0x03))
#define GC_GUARDED_PTR_P(p)   (((VALUE)p) & 0x01)

#define YARV_METHOD_NODE NODE_METHOD


/* defined? */
#define DEFINED_IVAR   INT2FIX(1)
#define DEFINED_GVAR   INT2FIX(2)
#define DEFINED_CVAR   INT2FIX(3)
#define DEFINED_CONST  INT2FIX(4)
#define DEFINED_METHOD INT2FIX(5)

/* for debug */
extern void vm_stack_dump_raw(yarv_thread_t *, yarv_control_frame_t *);
#define SDR() vm_stack_dump_raw(GET_THREAD(), GET_THREAD()->cfp)

#include "yarv.h"

#endif	// _YARVCORE_H_INCLUDED_



